//  -*- Mode: c++ -*-
#ifndef POINT
#define POINT

#include "acp.h"

using namespace acp;

class Point : public Object<PV2> {
public:
  Point () {}
  Point (double x, double y, bool perturb = true)
    : Object(PV2<double>(x, y), perturb) {}
  void print (bool newline = true) {
    PV2<double> p = getApproxMid();
    cerr << "(" << p.x << " " << p.y << ")";
    if (newline) cerr << endl;
  }
};

typedef vector<PTR<Point>> Points;

class XOrder : public Primitive {
  Point *a, *b;
  
  DeclareSign { return (b->get<N>().x - a->get<N>().x); }
public:
  XOrder (Point *a, Point *b) : a(a), b(b) {}
};

class YOrder : public Primitive {
  Point *a, *b;
  
  DeclareSign { return (b->get<N>().y - a->get<N>().y); }
public:
  YOrder (Point *a, Point *b) : a(a), b(b) {}
};

class CCW : public Primitive {
  Point *a, *b;
  
  DeclareSign { return a->get<N>().cross(b->get<N>()); }
public:
  CCW (Point *a, Point *b) : a(a), b(b) {}
};

class LeftTurn : public Primitive {
  Point *a, *b, *c;
  
  DeclareSign { return (c->get<N>() - b->get<N>()).cross(a->get<N>() - b->get<N>()); }
public:
  LeftTurn (Point *a, Point *b, Point *c) : a(a), b(b), c(c) {}
};

class Vector : public Point {
  PTR<Point> t, h;
  
  DeclareCalculate (PV2) { return h->get<N>() - t->get<N>(); }
public:
  Vector (Point *t, Point *h) : t(t), h(h) {}
};

class Normal : public Point {
  PTR<Point> t, h;

  DeclareCalculate (PV2) { 
    PV2<N> q = h->get<N>() - t->get<N>();
    return PV2<N>(- q.y, q.x); 
  }
public:
  Normal (Point *t, Point *h) : t(t), h(h) {}
};

class LineIntersection : public Point {
  PTR<Point> a, b, c, d;
  
  DeclareCalculate (PV2) {
    PV2<N> ap = a->get<N>(), cp = c->get<N>(), u = b->get<N>() - ap,
      v = d->get<N>() - cp;
    N k = (cp - ap).cross(v)/u.cross(v);
    return ap + k*u;
  }
public:
  LineIntersection (Point *a, Point *b, Point *c, Point *d) : a(a), b(b), c(c), d(d) {}
};

template<class N>
PV2<N> circumCenter (const PV2<N> &a, const PV2<N> &b, const PV2<N> &c)
{
  PV2<N> p = (a + b)/2, q = (b + c)/2, u(a.y - b.y, b.x - a.x), v(b.y - c.y, c.x - b.x);
  N k = (q - p).cross(v)/u.cross(v);
  return p + k*u;
}

class CircleCenterPoint : public Point {
  PTR<Point> a, b, c;

  DeclareCalculate (PV2) {
    return circumCenter(a->get<N>(), b->get<N>(), c->get<N>());
  }
public:
  CircleCenterPoint (Point *a, Point *b, Point *c) : a(a), b(b), c(c) {}
};

#endif
